searchData={"items":[{"type":"module","title":"Toolbox.Incident","doc":"Module extends Toolbox.Workflow and abstracts how regular incident behaves.\n\nWraps around `Toolbox.Workflow` and adds some additional callbacks to manage incident in asset\nmap. This module works very much like regular workflow, but some additional properties can be\nspecified. Contrary to a regular workflow, this automatically generates OAs to manage the incident\nin asset map and therefore syncs the general state of this workflow with the incident asset.\n\nStart by creating a definition (see `new/0`, `add_transition/2` and `build/1`) which describes the\nworkflow of the incident. Then, you can create a new instance based on that definition with\n`new_instance/7`.","ref":"Toolbox.Incident.html"},{"type":"function","title":"Toolbox.Incident.add_transition/2","doc":"Adds a new transition to incident workflow definition.\n\nIncident workflow is a finite state machine. Transition defines how workflow\nchanges its state from one to another and what must be done during the state change.\n\nIncident transition is defined by `params` parameter. `params` is a keyword list where\neach transition parameter has its key.\n Parameters may be either required or optional.\nSome parameters have predefined keys. These parameters are used to control workflow.\nThe developer of the scenario can add his own parameters to enrich function of the workflow\nwith features that the standard workflow module does not provide. These parameters can have\nany key except predefined ones.\n\nIf `add_transition/2` succeeds it returns a two level `Toolbox.Workflow.Transition` structure.\n - The first level contains parameters important for workflow control. Some parameters are copied\n   from `param` and some are computed.\n - The second level is located under attributes key. It contains parameters important\n   for workflow control and all parameters defined by scenario developer.\n\nTransition structure is passed, alongside with instance status and message to all callback\nfunctions. The scenario developer can evaluate them inside the function.\n\n- `from` - the state of the workflow from which the transition starts.\n  - required parameter\n- `to` - the state of the workflow after transition is completed.\n  - required parameter\n- `description_before` - transition from `from` state to `to` state generates an output action.\n  The output action contains simplified list of all possible states that can be reached from\n  `to` state. The list is stored in `body.attributes[\"future\"]`. Each entry of the list is\n  a map. The `description before` is assigned to `description` key of that map.\n  The `description` is intended to provide the human readable description of the conditions\n  leading to the particular state.\n  - required parameter\n- `description_after` - the aforementioned output action contains a list of all state changes\n  that the workflow has already passed through. That list is stored in `body.attributes[\"history']`.\n  Each entry of the list is a map containing details of the state. The `description_after`\n  is copied to the `\"description\"` key of the map. The `description` is intended to provide\n  human readable description about the past state change.\n  - required parameter\n- `severity` - an integer from 1 to 4. The severity of the incident after transition to `to` state.\n  - optional parameter\n- `when` - predicate used to select the transition which will be executed.\n  - optional parameter\n  - there can be multiple `when` definitions in list, all definitions are evaluated with\n    && boolean operator\n  - possible `when` definitions are:\n    - `{Module, function}`, where function accepts transition, instance and message as args,\n      returns boolean\n    - `{:timeout, timeout}`, where timeout is defined in milliseconds (transition is then\n      automatically executed when time elapses the specified value)\n    - `{:=, [path, to, state, key], value}` (transition is executed if the specified field of\n      state reaches the specified value)\n- `then` - callback used to update workflow instance state during transition execution.\n  - optional parameter\n  - there can be multiple then definitions in list, all definitions are executed in given order\n  - possible then definitions:\n    - `{Module, function}`, where function accepts transition, instance and message as args, and\n      returns `{:ok, state()}` to update the instance state\n- `side_effects` - callback used to generate side effects during transition execution.\n  - optional parameter\n  - there can be multiple side effect definitions in list, all definitions are executed\n    in given order.\n  - possible definitions:\n    - `{Module, function}`, where function accepts transition, instance and message as args, and\n      returns `{:ok, [OA | Msg | OtherSideEffect]}`\n- `update_history_entry` - callback used to modify transition execution history entry stored in\n  asset map. \n  - optional parameter\n  - there can be multiple definitions in list, all definitions are executed in given order\n  - this is usually used to interpolate description texts, or to add additional attributes to\n    history\n  - You can use this callback to set additional event parameters - `event_*` available in\n    `Runbox.Scenario.OutputAction.IncidentHistory` (the keys are strings in this case).\n  - possible definitions:\n    - `{Module, function}`, where function accepts history entry, transition, instance and message\n      as args and returns `{:ok, history_entry}`\n- `update_possible_transition` - callback used to modify possible future transitions stored in\n  asset map.\n  - optional parameter\n  - there can be multiple definitions in list, all definitions are executed in given order\n  - the callback is executed for each possible future transition\n  - note this only modifies the items of `future` attribute of the incident asset, this has no\n    effect on definition transitions\n  - possible definitions:\n    - `{Module, function}`, where function accepts possible transition, transition, instance\n      and message as args, and returns `{:ok, future_transition}`\n- `user_actions` - specifies all possible user actions from the target state\n  - a map of user actions that should be enabled once the transition is executed and incident is\n    in the target state.\n  - optional parameter\n  - user actions are automatically deleted if not present in the next transition\n  - keys are strings\n  - values are `{module, function}`, this specifies the function to be called to generate the user\n    action token (since tokens are not known in advance, they are generated by the specified\n    function)\n    - the function takes `transition, instance, message` as arguments and is expected to return\n      `{:ok, binary_token}` to register the user action token\n- `upsert_attributes` - specifies additional attributes to be added to the incident asset.\n  - optional parameter\n  - list of callbacks to compute the additional attributes\n  - each callback produces a map of additional attributes and this is merged into a single map\n    where the latter has priority over the former\n  - attributes cannot override attributes handled by this workflow incl. user actions, only other\n    attributes can be added\n  - possible definitions:\n    - `{Module, function}`, where function accepts transition, instance and message as args,\n      and should return `{:ok, attribute_map}`\n\n  Any other keys of the `params` parameter are put into the map under `attributes` key\n  of the `Transition` structure. They can be used inside callback functions for any purpose.\n  For example for debugging:","ref":"Toolbox.Incident.html#add_transition/2"},{"type":"function","title":"Code snippets - Toolbox.Incident.add_transition/2","doc":"def definition do\n        Incident.new()\n        |> Incident.add_transition(\n          from: \"detected\",\n          to: \"rare\",\n          when: [{__MODULE__, :rare_message_arrived?}],\n          description_before: \"Other messages will arrive\",\n          description_after: \"Rare message arrived\",\n          update_history_entry: [{__MODULE__, :update_history_entry}],\n          my_debug_key: \"This should be a very rare transition\"\n          )\n          ...\n        |> Incident.build()\n      end\n\n      def update_history_entry(history, transition, _workflow_instance, message) do\n        Logger.info(transition.attributes.my_debug_key)\n        ...\n        {:ok, new_history}\n      end\n\nWhen a message is evaluated the callbacks above are run in the following order.\n  1. `when` callbacks are evaluated to see if the current transition is ready to be executed. If\n     not the next transition is tried.\n  2. `then` callbacks are evaluated to update the instance state.\n  3. `user_actions` is evaluated, all callbacks specified inside are executed and all user action\n     tokens are calculated.\n  4. `update_history_entry` callbacks are evaluated to update the new history entry\n  5. `update_possible_transition` callbacks are evaluated to update the new possible future\n     transitions\n  6. `upsert_attributes` callbacks are evaluated to gather additional attributes\n  7. `side_effects` callbacks are evaluated to acquire the list of all additional output actions\n\nAll text bearing attributes (such as `subject`, `description_before`,\n`description_after`) have access to incident metadata dictionary. This dictionary contains these,\nkeys:\n  - `type` - incident type\n  - `id` - incident id\n  - `transition` - transition attributes dictionary containing `from`, `to`, `severity` keys\n  - `state` - dictionary containing user defined state\n  - `message` - altworx message which triggered given transition\n\nMetadata can be accessed via interpolation defined as `{{}}`, e.g. `{{state.foo.bar}}`,\n`{{message.body.foo}}`. There is also an option to reference assets in those attributes, see\n`t:Runbox.Scenario.OutputAction.interpolable/0`.","ref":"Toolbox.Incident.html#add_transition/2-code-snippets"},{"type":"function","title":"Toolbox.Incident.append_create_incident_output_actions/3","doc":"","ref":"Toolbox.Incident.html#append_create_incident_output_actions/3"},{"type":"function","title":"Toolbox.Incident.append_update_incident_output_actions/3","doc":"","ref":"Toolbox.Incident.html#append_update_incident_output_actions/3"},{"type":"function","title":"Toolbox.Incident.build/1","doc":"","ref":"Toolbox.Incident.html#build/1"},{"type":"function","title":"Toolbox.Incident.handle_message/3","doc":"Uses given incident workflow definition and message to update state of given instance.\n\nIf no configured workflow transition matches, nothing will happen = instance state will remain the\nsame.\n\nOrder of callback execution:\n1. when definitions of transitions in definition order\n2. then definitions of matching transition\n3. update history entry definitions of matching transition\n3. update possible transition definitions of matching transition\n5. side effects definitions of matching transition","ref":"Toolbox.Incident.html#handle_message/3"},{"type":"function","title":"Toolbox.Incident.new/0","doc":"Creates new blank incident workflow definition","ref":"Toolbox.Incident.html#new/0"},{"type":"function","title":"Toolbox.Incident.new_instance/7","doc":"Creates new incident instance for given workflow.\n\n`params` can be used to specify additional transition-like parameters. `actors` parameter is evaluated\nonly when `new_instance/7` is called and is not stored in the instance. It is a listing of actors used\nwhen an incident is created. Its format is a list of maps where `id` and `type` are mandatory keys.\n`id` and `type` references an asset.\nFor other available options see `add_transition/2`.","ref":"Toolbox.Incident.html#new_instance/7"},{"type":"function","title":"Toolbox.Incident.update_incident_history_entry/4","doc":"","ref":"Toolbox.Incident.html#update_incident_history_entry/4"},{"type":"function","title":"Toolbox.Incident.update_incident_possible_transition/4","doc":"","ref":"Toolbox.Incident.html#update_incident_possible_transition/4"},{"type":"function","title":"Toolbox.Incident.update_incident_state/3","doc":"","ref":"Toolbox.Incident.html#update_incident_state/3"},{"type":"function","title":"Toolbox.Incident.update_user_actions/3","doc":"","ref":"Toolbox.Incident.html#update_user_actions/3"},{"type":"function","title":"Toolbox.Incident.validate_upsert_attributes/1","doc":"","ref":"Toolbox.Incident.html#validate_upsert_attributes/1"},{"type":"module","title":"Toolbox.Incident.IdGenerator","doc":"Module provides functions to generate incident id.","ref":"Toolbox.Incident.IdGenerator.html"},{"type":"function","title":"Toolbox.Incident.IdGenerator.generate/3","doc":"Generates full incident ID based on incident type, timestamp and serialization vector.\n\nSerialization vector is list of relevant attributes which together with type and timestamp\nuniquely identify given incident.\n\nNote including maps in serialization vector increases a risk the ID will change on new major OTP\nversion. If you strongly prefer long-term stability avoid maps. This uses\n`:erlang.term_to_binary` under the hood.\n\nThis function is deprecated as it returns the full ID that includes type. This is no longer very\nuseful as Output Actions take type and ID separately. Use `generate_id/3` instead.","ref":"Toolbox.Incident.IdGenerator.html#generate/3"},{"type":"function","title":"Examples - Toolbox.Incident.IdGenerator.generate/3","doc":"iex> Toolbox.Incident.IdGenerator.generate(\"/incident/type\", 0123456789, [\"/asset/type/1\"])\n      \"/incident/type/f4c3877a-d50b-3a16-b6ec-3ebb2b308d40\"","ref":"Toolbox.Incident.IdGenerator.html#generate/3-examples"},{"type":"function","title":"Toolbox.Incident.IdGenerator.generate_id/3","doc":"Generates incident ID based on incident type, timestamp and serialization vector.\n\nSerialization vector is list of relevant attributes which together with type and timestamp\nuniquely identify given incident.\n\nNote including maps in serialization vector increases a risk the ID will change on new major OTP\nversion. If you strongly prefer long-term stability avoid maps. This uses\n`:erlang.term_to_binary` under the hood.\n\nThis function is compatible with deprecated `generate/3`, you can replace it with this and the IDs\nwill not change.","ref":"Toolbox.Incident.IdGenerator.html#generate_id/3"},{"type":"function","title":"Examples - Toolbox.Incident.IdGenerator.generate_id/3","doc":"iex> Toolbox.Incident.IdGenerator.generate_id(\"/incident/type\", 0123456789, [\"/asset/type/1\"])\n      \"f4c3877a-d50b-3a16-b6ec-3ebb2b308d40\"","ref":"Toolbox.Incident.IdGenerator.html#generate_id/3-examples"},{"type":"module","title":"Toolbox.Random","doc":"Functions for working with deterministic (pseudo)randomness in scenarios.\n\nIn some use cases, scenarios need to behave “randomly”, for example, when\na scenario serves as a demo events generator. Scenarios are by design\ndeterministic which means that they can only behave randomly in the sense\nthat they make certain decisions “on their own” (without the programmer\nexplicitly choosing the outcome) but every time a scenario run is repeated,\nthe same “random” decisions are made. To make this possible, the seed of\nthe random number generator has to be always initialized to the same value\nand managed in the way that it is preserved throughout the run. And that\nis what this module helps with.\n\nThe module wraps around Erlang `:rand` module for random number generation.\nFunctions in the `:rand` module use process dictionary for storing the\nrandom number generator state (i.e. the seed). To be able to restore a run,\nthe random number generator state is also kept in the state of a unit in\nan opaque field called `__rand_seed`.\n\nThere are three steps (functions of this module) in order to use randomness\nin scenarios. First, you have to initialize the unit that uses randomness\nwith `init/2`. And then, every time before you want to use randomness, you\nhave to call `load/1` with the particular unit. And then store the new\nseed with `save/1`. An example is given below. See the documentation for\neach of the functions for more detailed information.","ref":"Toolbox.Random.html"},{"type":"module","title":"Example - Toolbox.Random","doc":"defmodule Scenarios.MyScenario.MyTemplate\n      @behaviour Runbox.Scenario.Template.StageBased\n\n      alias Toolbox.Random\n\n      def instances do\n        # singleton unit\n        [{ :unit,\"generator\", %{} }]\n      end\n\n      def init(_start_from, unit) do\n        unit = Random.init(unit, 293732)\n        {:ok, [], unit}\n      end\n\n      def handle_message(msg, unit) do\n        Random.load(unit)\n        ... do something that uses :rand calls\n        ... (e.g. :rand.uniform/0 or Enum.random/1)\n        {:noreply, oas, Random.save(unit)}\n      end\n    end","ref":"Toolbox.Random.html#module-example"},{"type":"macro","title":"Toolbox.Random.init/1","doc":"Shortcut for `init/2` called with `__MODULE__` as the seed.\n\nIt is a bit shorter, but on the other hand you have to `require Toolbox.Random`\nto be able to use this macro. Beware that the generated values will differ when\nyou rename your template module.","ref":"Toolbox.Random.html#init/1"},{"type":"function","title":"Toolbox.Random.init/2","doc":"Initializes a unit for random number generation.\n\nYou need to provide an initial seed which may be any term, in particular\nany integer, string or atom. The value doesn't matter (it only affects the\nrandomly generated values) as long as it is constant. Typically, you would\nuse this function inside the `init/2` callback of your template.\n\nIn case you are using a singleton unit, you may, for example, hardcode an\ninteger here or use `__MODULE__` as the initial seed:\n\n    def init(_start_from, unit) do\n      unit = Random.init(unit, __MODULE__)\n      {:ok, [], unit}\n    end\n\nIf you have multiple units (e.g. unit for each person), you may initialize\neach unit with a different seed (e.g. the ID of that person):\n\n    def init(_start_from, unit) do\n      unit = Random.init(unit, unit.attributes.id)\n      {:ok, [], unit}\n    end\n\nThis function saves the initial seed into the unit's state under\nan opaque field called `__rand_seed` and return the modified unit.\nTherefore, to use randomness, the unit's state has to be a map.\n\nThe function also sets the seed in the process dictionary so you\ncan start using randomness right away. But you will probably want\nto do so later in the `handle_message/2` callback.","ref":"Toolbox.Random.html#init/2"},{"type":"function","title":"Toolbox.Random.load/1","doc":"Loads a seed from the unit's state into the process dictionary.\n\nTypically, you would call this function in the beginning of a `handle_message/2`\ncallback that deals with randomness. It is sufficient to call it once per invocation\nof the callback.","ref":"Toolbox.Random.html#load/1"},{"type":"function","title":"Example - Toolbox.Random.load/1","doc":"def handle_message(msg, unit) do\n      Random.load(unit)\n\n      person = Enum.random(unit.state.people)\n      room = Enum.random(unit.state.rooms)\n      # generate an event that $person went into $room\n\n      {:noreply, oas, Random.save(unit)}\n    end","ref":"Toolbox.Random.html#load/1-example"},{"type":"function","title":"Toolbox.Random.save/1","doc":"Stores the random generator state in the unit's state.\n\nTypically, you would use this function after generating random values\n(and thus modifying the generator's seed) in the `handle_message/2` callback.\n\nThe function returns the modified unit.","ref":"Toolbox.Random.html#save/1"},{"type":"type","title":"Toolbox.Random.input_seed/0","doc":"","ref":"Toolbox.Random.html#t:input_seed/0"},{"type":"type","title":"Toolbox.Random.seed/0","doc":"","ref":"Toolbox.Random.html#t:seed/0"},{"type":"behaviour","title":"Toolbox.Report","doc":"Behaviour module for notification reports.\n\nThis behaviour encapsulates notifications which behave like reports = notifications are sent\nperiodically.","ref":"Toolbox.Report.html"},{"type":"behaviour","title":"Example - Toolbox.Report","doc":"```\ndefmodule Scenarios.Template.Reporter do\n  @callback Runbox.Scenario.Template.StageBased\n\n  alias Crontab.CronExpression\n  alias Scenarios.Template.Reporter.DailyReport\n  alias Toolbox.Message\n  alias Toolbox.Report\n  alias Toolbox.Runtime.Stage.Unit\n\n  import Crontab.CronExpression\n\n  @impl true\n  def init(init_timestamp, %Unit{} = unit) do\n    # report schedule defined as crontab expression\n    report_schedule = ~e[* * * * *]\n\n    # timezone used to display localized timestamp in notification\n    timezone = \"Europe/Prague\"\n\n    # args to init report state\n    init_arg = :ok\n\n    {:ok, outputs, report_state} =\n      Report.init(DailyReport, init_timestamp, report_schedule, timezone, init_arg)\n\n    {:ok, outputs, %Unit{unit | state: %{report_state: report_state}}}\n  end\n\n  @impl true\n  def handle_message(%Message{} = msg, %Unit{} = unit) do\n    {:ok, outputs, report_state} = Report.process_message(unit.state.report_state, msg)\n    {:reply, outputs, %Unit{unit | state: %{unit.state | report_state: report_state}}}\n  end\nend\n\ndefmodule Scenarios.Template.Reporter.DailyReport do\n  @callback Toolbox.Report\n\n  alias Runbox.Scenario.OutputAction\n  alias Toolbox.Message\n\n  @impl true\n  def init(:ok) do\n    {:ok, [], %{incoming_messages: []}}\n  end\n\n  @impl true\n  def handle_process_message(%{} = state, %Message{} = msg) do\n    {:ok, [], %{incoming_messages: [msg | state.incoming_messages]}}\n  end\n\n  @impl true\n  def handle_publish_report(%{} = state, publish_timestamp, last_publish_timestamp) do\n    report_notification =\n      %OutputAction.Notification{\n        type: :daily_report,\n        primary_asset: nil,\n        data: %{\n          \"incoming_messages\" => state.incoming_messages,\n          \"publish_time\" => publish_timestamp,\n          \"last_publish_time\" => last_timestamp\n        }\n      }\n\n    {:ok, [report_notification], %{incoming_messages: []}}\n  end\nend\n```","ref":"Toolbox.Report.html#module-example"},{"type":"callback","title":"Toolbox.Report.handle_process_message/2","doc":"","ref":"Toolbox.Report.html#c:handle_process_message/2"},{"type":"callback","title":"Toolbox.Report.handle_publish_report/3","doc":"","ref":"Toolbox.Report.html#c:handle_publish_report/3"},{"type":"callback","title":"Toolbox.Report.init/1","doc":"","ref":"Toolbox.Report.html#c:init/1"},{"type":"function","title":"Toolbox.Report.init/5","doc":"Initializes report","ref":"Toolbox.Report.html#init/5"},{"type":"function","title":"Toolbox.Report.process_message/2","doc":"Processes report message","ref":"Toolbox.Report.html#process_message/2"},{"type":"function","title":"Toolbox.Report.resolve_ambiguous_datetime/1","doc":"Resolves date time into non-ambiguous DateTime.\n\nIf AmbiguousDateTime is given, it picks one version. If any other non-ambiguous date time is\ngiven it is returned as is.","ref":"Toolbox.Report.html#resolve_ambiguous_datetime/1"},{"type":"opaque","title":"Toolbox.Report.external_state/0","doc":"State that the `Toolbox.Report` operates on; should be stored in a unit's state.","ref":"Toolbox.Report.html#t:external_state/0"},{"type":"type","title":"Toolbox.Report.report_state/0","doc":"User state where a module implementing `Toolbox.Report` can store anything it wants.","ref":"Toolbox.Report.html#t:report_state/0"},{"type":"type","title":"Toolbox.Report.timestamp/0","doc":"","ref":"Toolbox.Report.html#t:timestamp/0"},{"type":"module","title":"Toolbox.Utils.Enum","doc":"A set of utility functions for `Enumerable`s.","ref":"Toolbox.Utils.Enum.html"},{"type":"function","title":"Toolbox.Utils.Enum.each_while_ok/2","doc":"Returns `:ok` if `fun` applied to every element of `enumerable` returns `:ok`.\nOtherwise it returns the first error occurred during `fun` call.","ref":"Toolbox.Utils.Enum.html#each_while_ok/2"},{"type":"function","title":"Example - Toolbox.Utils.Enum.each_while_ok/2","doc":"iex> each_while_ok([], fn _ -> :not_called_at_all end)\n    :ok\n\n    iex> each_while_ok([1, 2, 3], fn _ -> :ok end)\n    :ok\n\n    iex> each_while_ok([1, 2, 3], &(if &1 == 2, do: {:error, &1}, else: :ok))\n    {:error, 2}","ref":"Toolbox.Utils.Enum.html#each_while_ok/2-example"},{"type":"function","title":"Toolbox.Utils.Enum.flatten_levels/2","doc":"Flattens the enum by the specified amount of levels.\n\nFlattens only the specified amount of levels, contrary to the standard `List.flatten/2` which\nflattens the list recursively. Handy if you e.g. want to flatten only the first level.\n\nImportantly, this operation maintains order of the elements in the list.","ref":"Toolbox.Utils.Enum.html#flatten_levels/2"},{"type":"function","title":"Examples - Toolbox.Utils.Enum.flatten_levels/2","doc":"iex> flatten_levels([[:hello, :team], [\"this\", \"is\"], [\"now\", [\"flat\", [\"list\"]]]], 1)\n    [:hello, :team, \"this\", \"is\", \"now\", [\"flat\", [\"list\"]]]\n\n    iex> flatten_levels([[:hello, :team], [\"this\", \"is\"], [\"now\", [\"flat\", [\"list\"]]]], 2)\n    [:hello, :team, \"this\", \"is\", \"now\", \"flat\", [\"list\"]]\n\n    iex> flatten_levels([[:hello, :team], [\"this\", \"is\"], [\"now\", [\"flat\", [\"list\"]]]], 3)\n    [:hello, :team, \"this\", \"is\", \"now\", \"flat\", \"list\"]\n\n    iex> flatten_levels([[:hello, :team], [\"this\", \"is\"], [\"now\", [\"flat\", [\"list\"]]]], 4)\n    [:hello, :team, \"this\", \"is\", \"now\", \"flat\", \"list\"]\n\n    iex> flatten_levels([[:hello, :team], [\"this\", \"is\"], [\"now\", [\"flat\", [\"list\"]]]], 0)\n    [[:hello, :team], [\"this\", \"is\"], [\"now\", [\"flat\", [\"list\"]]]]","ref":"Toolbox.Utils.Enum.html#flatten_levels/2-examples"},{"type":"function","title":"Toolbox.Utils.Enum.map_while_ok/2","doc":"Returns `{:ok, list_of_values}` if `fun` applied to every element of\n`enumerable` returns `{:ok, value}`. Otherwise it returns the first\nerror occurred during `fun` call.","ref":"Toolbox.Utils.Enum.html#map_while_ok/2"},{"type":"function","title":"Example - Toolbox.Utils.Enum.map_while_ok/2","doc":"iex> map_while_ok([], fn _ -> :not_called_at_all end)\n    {:ok, []}\n\n    iex> map_while_ok([1, 2, 3], &({:ok, &1 * 2}))\n    {:ok, [2, 4, 6]}\n\n    iex> map_while_ok([1, 2, 3], &(if &1 == 2, do: {:error, &1 * 2}, else: {:ok, &1 * 2}))\n    {:error, 4}","ref":"Toolbox.Utils.Enum.html#map_while_ok/2-example"},{"type":"function","title":"Toolbox.Utils.Enum.reduce_while_ok/3","doc":"Invokes `fun` for each element in the `enumerable` with the accumulator.\nIf every `fun` call returns `{:ok, new_acc}`, functions returns `{:ok, last_acc}`.\nOtherwise it returns the first error occurred during `fun` call.","ref":"Toolbox.Utils.Enum.html#reduce_while_ok/3"},{"type":"function","title":"Example - Toolbox.Utils.Enum.reduce_while_ok/3","doc":"iex> reduce_while_ok([], :initial, fn _, _ -> :not_called_at_all end)\n    {:ok, :initial}\n\n    iex> reduce_while_ok([1, 2, 3], 0, &({:ok, &1 + &2}))\n    {:ok, 6}\n\n    iex> reduce_while_ok([1, 2, 3], 0, &(if &1 == 2, do: {:error, &1 + &2}, else: {:ok, &1 + &2}))\n    {:error, 3}","ref":"Toolbox.Utils.Enum.html#reduce_while_ok/3-example"},{"type":"module","title":"Toolbox.Utils.Map","doc":"A set of utility functions for maps.","ref":"Toolbox.Utils.Map.html"},{"type":"function","title":"Toolbox.Utils.Map.deep_create/2","doc":"Creates deep map with `value` on key represented by `path`. If `path` is empty,\nit returns given `value`.","ref":"Toolbox.Utils.Map.html#deep_create/2"},{"type":"function","title":"Examples - Toolbox.Utils.Map.deep_create/2","doc":"iex> Toolbox.Utils.Map.deep_create([], :some_value)\n    :some_value\n\n    iex> Toolbox.Utils.Map.deep_create([\"a\", :b], :other_value)\n    %{\"a\" => %{:b => :other_value}}","ref":"Toolbox.Utils.Map.html#deep_create/2-examples"},{"type":"function","title":"Toolbox.Utils.Map.deep_merge/2","doc":"Deeply updates `left` map by `right` map.\n\nExisting keys from `left` map not contained in `right` map are unchanged. New keys from `right`\nmap are added. If the key is present in both maps and both values are maps, maps are deeply\nmerges. If the key is present in both maps and value of one of them is not a map, value of key\nin `left` map is replaced by value in `right` map.","ref":"Toolbox.Utils.Map.html#deep_merge/2"},{"type":"function","title":"Example - Toolbox.Utils.Map.deep_merge/2","doc":"iex> Toolbox.Utils.Map.deep_merge(%{a: 1}, %{b: 2})\n    %{a: 1, b: 2}\n\n    iex> Toolbox.Utils.Map.deep_merge(%{a: %{b: 1}}, %{a: %{c: 3}})\n    %{a: %{b: 1, c: 3}}\n\n    iex> Toolbox.Utils.Map.deep_merge(%{a: %{b: %{c: 1}}}, %{a: %{b: %{d: 2}}})\n    %{a: %{b: %{c: 1, d: 2}}}\n\n    iex> Toolbox.Utils.Map.deep_merge(%{a: 1}, %{a: %{b: 2}})\n    %{a: %{b:  2}}\n\n    iex> Toolbox.Utils.Map.deep_merge(%{a: %{b: 1}}, %{a: 2})\n    %{a: 2}\n\n    iex> Toolbox.Utils.Map.deep_merge(%{a: 1}, %{a: 2})\n    %{a: 2}","ref":"Toolbox.Utils.Map.html#deep_merge/2-example"},{"type":"function","title":"Toolbox.Utils.Map.deep_remove/2","doc":"Removes key represented as path (list of keys) deeply in map if it exists.","ref":"Toolbox.Utils.Map.html#deep_remove/2"},{"type":"function","title":"Example - Toolbox.Utils.Map.deep_remove/2","doc":"iex> Toolbox.Utils.Map.deep_remove(%{a: 1, b: 2}, [])\n    %{a: 1, b: 2}\n\n    iex> Toolbox.Utils.Map.deep_remove(%{a: 1, b: 2}, [:a])\n    %{b: 2}\n\n    iex> Toolbox.Utils.Map.deep_remove(%{a: 1, b: 2}, [:a, :aa])\n    %{a: 1, b: 2}\n\n    iex> Toolbox.Utils.Map.deep_remove(%{a: %{aa: 1, ab: 2}, b: 2}, [:a, :aa])\n    %{a: %{ab: 2}, b: 2}\n\n    iex> Toolbox.Utils.Map.deep_remove(%{a: %{aa: 1, ab: 2}, b: 2}, [:a, :aa, :aaa])\n    %{a: %{aa: 1, ab: 2}, b: 2}","ref":"Toolbox.Utils.Map.html#deep_remove/2-example"},{"type":"function","title":"Toolbox.Utils.Map.from_flat_map/1","doc":"Converts map from flat to nested map.\n\nThe input map must be flat with array keys, which represent the path to the respective values.\nThe corresponding nested map is then constructed.\n\nAlso see `to_flat_map/1` which performs the reverse operation.","ref":"Toolbox.Utils.Map.html#from_flat_map/1"},{"type":"function","title":"Examples - Toolbox.Utils.Map.from_flat_map/1","doc":"iex> Toolbox.Utils.Map.from_flat_map(%{\n    ...>   [\"contact\", \"name\"] => \"Jored\",\n    ...>   [\"contact\", \"address\", \"city\"] => \"Lrno\",\n    ...>   [\"active\", \"ui\"] => true,\n    ...>   [\"active\", \"backend\"] => false\n    ...> })\n    %{\n      \"contact\" => %{\n        \"name\" => \"Jored\",\n        \"address\" => %{\n          \"city\" => \"Lrno\"\n        }\n      },\n      \"active\" => %{\n        \"ui\" => true,\n        \"backend\" => false\n      }\n    }\n\n    iex> Toolbox.Utils.Map.from_flat_map(%{\n    ...>   [:very, :very, :very, :very, :very, :very, :deep] => \"object\",\n    ...>   [:very, :very, :very, :very, :very, :very, :nested] => \"map\"\n    ...> })\n    %{\n      very: %{\n        very: %{\n          very: %{\n            very: %{\n              very: %{\n                very: %{\n                  deep: \"object\",\n                  nested: \"map\"\n                }\n              }\n            }\n          }\n        }\n      }\n    }\n\n    iex> Toolbox.Utils.Map.from_flat_map(%{})\n    %{}\n\n    iex> Toolbox.Utils.Map.from_flat_map(Toolbox.Utils.Map.to_flat_map(%{\n    ...>   \"contact\" => %{\n    ...>     \"name\" => \"Jonas\",\n    ...>     \"address\" => %{\"city\" => \"Brnp\"}},\n    ...>     \"favorites\" => %{\n    ...>       \"food\" => %{\n    ...>         1 => [\"cornflakes\", \"corn\"],\n    ...>         5 => [\"pizza\"],\n    ...>         8 => [\"pasta\"]\n    ...>       }\n    ...>     }\n    ...>   }\n    ...> ))\n    %{\n      \"contact\" => %{\n        \"name\" => \"Jonas\",\n        \"address\" => %{\"city\" => \"Brnp\"}},\n        \"favorites\" => %{\n          \"food\" => %{\n            1 => [\"cornflakes\", \"corn\"],\n            5 => [\"pizza\"],\n            8 => [\"pasta\"]\n        }\n      }\n    }","ref":"Toolbox.Utils.Map.html#from_flat_map/1-examples"},{"type":"function","title":"Toolbox.Utils.Map.get_path/3","doc":"Follows the `path` in the (deep) map, defaulting to `default` if some of the\nobjects on the path is missing.","ref":"Toolbox.Utils.Map.html#get_path/3"},{"type":"function","title":"Example - Toolbox.Utils.Map.get_path/3","doc":"iex> object = %{\"some\" => %{\"deep\" => %{\"object\" => \"data\"}}}\n    ...> Toolbox.Utils.Map.get_path(object, [\"some\", \"deep\", \"object\"])\n    \"data\"\n    iex> Toolbox.Utils.Map.get_path(object, [\"non-existent\"], \"default\")\n    \"default\"","ref":"Toolbox.Utils.Map.html#get_path/3-example"},{"type":"function","title":"Toolbox.Utils.Map.json_paths/1","doc":"Extracts all (json-like) paths in the map.\n\nGiven any artitrarily nested map, the function returns all paths the map contains.\n\nThe output is **not** sorted in a predictable way.","ref":"Toolbox.Utils.Map.html#json_paths/1"},{"type":"function","title":"Examples - Toolbox.Utils.Map.json_paths/1","doc":"iex> Toolbox.Utils.Map.json_paths(%{\n    ...>    \"contact\" => %{\n    ...>      \"name\" => \"Trest\",\n    ...>      \"phone\" => 605554171,\n    ...>      \"address\" => %{\"city\" => \"Brno\"}\n    ...>    },\n    ...>    \"active\" => false\n    ...> }) |> Enum.sort()\n    [\n      [\"active\"],\n      [\"contact\", \"address\", \"city\"],\n      [\"contact\", \"name\"],\n      [\"contact\", \"phone\"],\n    ]\n\n    iex> Toolbox.Utils.Map.json_paths(%{\"a\" => %{1 => %{%{\"cool\" => \"stuff\"} => %{atom: :work_too}}}})\n    [\n      [\"a\", 1, %{\"cool\" => \"stuff\"}, :atom]\n    ]\n\n    iex> Toolbox.Utils.Map.json_paths(%{})\n    []","ref":"Toolbox.Utils.Map.html#json_paths/1-examples"},{"type":"function","title":"Toolbox.Utils.Map.subtract/2","doc":"Subtracts map from another map, the leaf-paths from the second map are\nremoved from the first.","ref":"Toolbox.Utils.Map.html#subtract/2"},{"type":"function","title":"Examples - Toolbox.Utils.Map.subtract/2","doc":"iex> Toolbox.Utils.Map.subtract(%{a: 1, b: 2}, %{a: true})\n    %{b: 2}\n\n    iex> Toolbox.Utils.Map.subtract(%{a: 1}, %{a: %{b: %{c: true}}})\n    %{a: 1}\n\n    iex> Toolbox.Utils.Map.subtract(\n    ...>   %{a: %{b: %{c: \"1\"}, d: \"2\"}, e: \"3\"},\n    ...>   %{a: %{b: %{c: true}}, e: true}\n    ...> )\n    %{a: %{b: %{}, d: \"2\"}}\n\n    iex> Toolbox.Utils.Map.subtract(\n    ...>   %{\"a\" => 1, \"b\" => %{\"c\" => true}},\n    ...>   %{\"b\" => %{\"c\" => true}}\n    ...> )\n    %{\"a\" => 1, \"b\" => %{}}\n\n    iex> Toolbox.Utils.Map.subtract(%{a: 1, b: 2}, %{c: 2, d: 3})\n    %{a: 1, b: 2}\n\n    iex> Toolbox.Utils.Map.subtract(%{a: %{b: 1}}, %{a: true})\n    %{}","ref":"Toolbox.Utils.Map.html#subtract/2-examples"},{"type":"function","title":"Toolbox.Utils.Map.to_flat_map/1","doc":"Converts regular nested map to flat map.\n\nThe input map can be any map - arbitrarily nested, the output is a flat map where keys are arrays\nrepresenting the path to the respective value.\n\nAlso see `from_flat_map/1` which performs the reverse operation.","ref":"Toolbox.Utils.Map.html#to_flat_map/1"},{"type":"function","title":"Examples - Toolbox.Utils.Map.to_flat_map/1","doc":"iex> Toolbox.Utils.Map.to_flat_map(%{\n    ...>   \"contact\" => %{\n    ...>     \"name\" => \"Kert\",\n    ...>     \"address\" => %{\"city\" => \"Krno\"}\n    ...>   },\n    ...>   \"status\" => %{\"ui\" => %{\"active\" => true},\n    ...>   \"backend\" => %{\"fluent\" => false}}\n    ...> })\n    %{\n      [\"contact\", \"name\"] => \"Kert\",\n      [\"contact\", \"address\", \"city\"] => \"Krno\",\n      [\"status\", \"ui\", \"active\"] => true,\n      [\"status\", \"backend\", \"fluent\"] => false\n    }\n\n    iex> Toolbox.Utils.Map.to_flat_map(%{good: %{1 => :nice}, very: %{deep: %{%{ish: :map} => :askey}}})\n    %{\n      [:good, 1] => :nice,\n      [:very, :deep, %{ish: :map}] => :askey\n    }\n\n    iex> Toolbox.Utils.Map.to_flat_map(%{\"regular\" => \"map\"})\n    %{\n      [\"regular\"] => \"map\"\n    }\n\n    iex> Toolbox.Utils.Map.to_flat_map(%{})\n    %{}","ref":"Toolbox.Utils.Map.html#to_flat_map/1-examples"},{"type":"module","title":"Toolbox.Workflow","doc":"Workflow is an abstraction of a state machine designed to use with Altworx scenarios.","ref":"Toolbox.Workflow.html"},{"type":"module","title":"Overview - Toolbox.Workflow","doc":"A workflow essentially describes a “blueprint” for a state machine. It specifies what\npossible states of the machine are, when does it move from one state to another,\nand what happens when such a transition is performed.\n\nA concrete state machine that works according to the blueprint is called a *workflow\ninstance* and is represented by the `Toolbox.Workflow.Instance` struct. It specifies\nwhat the current status and state are, what statuses it has gone through so far, and\nso on. See `t:Toolbox.Workflow.Instance.t/0` for more information.\n\nNote that what is normally referred to as *states* are in workflows called *statuses*.\nStatuses are from a finite set and are given in advance (e.g., a CD player can have\nthree statuses: `empty`, `idle`, and `playing`). *State*, on the other hand, is an\nadditional context enriching status with other information. State is not listed upfront\nand is not limited in size. For example, state of a CD player can be the artist and\nalbum of the CD that is currently playing.\n\nIn a workflow, statuses are defined by declared transitions. That is, the set of statuses\nis defined by all the statuses that are mentioned in the definition of transitions.\nStatuses are therefore a part of the workflow definition and cannot be changed later.\nState is defined when a new workflow instance is created. State can then be modified\nduring the life-cycle of the instance using `then` callbacks.\n\nWorkflow instances consume messages received by scenarios and produce output actions.\nThey are used from inside scenario units which feed them with messages, return produced\noutput actions, and store their state (i.e., their instance struct) in their own state.\nAs units are event-driven and can only react to messages, so are workflow instances and\nthe only time when a transition may be performed (or when a workflow instance may be\ncreated) is when a message is received.\n\nMost of the time, workflows are used via their wrapper for incident workflows\n(`Toolbox.Incident`). However, they can be used on their own as well.","ref":"Toolbox.Workflow.html#module-overview"},{"type":"module","title":"Example - Toolbox.Workflow","doc":"Here is a simple workflow for a CD player that has statuses and transitions as defined\nin the diagram below and that creates an event when a CD finishes.\nIt operates on state that can be described with the following type:\n`%{cd: %{album: String.t(), artist: String.t()} | nil}`.\n\n```mermaid\nstateDiagram-v2\n  direction LR\n  empty --> idle: insert CD\n  idle --> playing: play\n  playing --> idle: stop\n  playing --> idle: end of CD\n  playing --> empty: eject\n  idle --> empty: eject\n```\n\n    defmodule CdPlayerWorkflow do\n      alias Toolbox.Workflow\n      alias Toolbox.Message\n      alias Runbox.Scenario.OutputAction, as: OA\n\n      @cd_length :timer.minutes(74)\n\n      def definition do\n        Workflow.new()\n        |> Workflow.add_transition(from: \"empty\", to: \"idle\", when: {__MODULE__, :insert_cd_msg},\n                                   then: {__MODULE__, :put_cd_to_state})\n        |> Workflow.add_transition(from: \"idle\", to: \"playing\", when: {__MODULE__, :play_msg})\n        |> Workflow.add_transition(from: \"playing\", to: \"idle\", when: {__MODULE__, :stop_msg})\n        |> Workflow.add_transition(from: \"playing\", to: \"idle\", when: {:timeout, @cd_length},\n                                   side_effects: {__MODULE__, :create_played_event})\n        |> Workflow.add_transition(from: \"playing\", to: \"empty\", when: {__MODULE__, :eject_msg},\n                                   then: {__MODULE__, :remove_cd_from_state})\n        |> Workflow.add_transition(from: \"idle\", to: \"empty\", when: {__MODULE__, :eject_msg},\n                                   then: {__MODULE__, :remove_cd_from_state})\n        |> Workflow.build()\n      end\n\n      def insert_cd_msg(_tran, _inst, %Message{type: type}), do: type == :insert_cd\n      def eject_msg(_tran, _inst, %Message{type: type}), do: type == :eject\n      def play_msg(_tran, _inst, %Message{type: type}), do: type == :play\n      def stop_msg(_tran, _inst, %Message{type: type}), do: type == :stop\n\n      def put_cd_to_state(_tran, inst, %Message{type: :insert_cd, body: body}) do\n        cd = %{album: body.album, artist: body.artist}\n        state = %{inst.state | cd: cd}\n        {:ok, state}\n      end\n\n      def remove_cd_from_state(_tran, inst, _msg) do\n        state = %{inst.state | cd: nil}\n        {:ok, state}\n      end\n\n      def create_played_event(_tran, inst, msg) do\n        oa = %OA.Event{\n          type: \"played_cd\",\n          template: \"${actors.player} played ${params.album} by ${params.artist}\",\n          actors: %{\"player\" => %{asset_type: inst.type, asset_id: inst.id, name: nil}},\n          params: %{\"album\" => inst.state.cd.album, \"artist\" => inst.state.cd.artist},\n          origin_messages: [msg.origin]\n        }\n        {:ok, [oa]}\n      end\n    end\n\nAn instance of this workflow may be then created like this:\n\n    {:ok, wf} = CdPlayerWorkflow.definition()\n    {:ok, oas, inst} = Workflow.new_instance(wf, \"empty\", \"cd_player\", \"Sony D-50\", %{cd: nil}, msg)\n\nAfterwards, when a message comes to your unit, you can then call `handle_message/3` to process\nthe message and get an updated instance together with a list of produced output actions:\n\n    {:ok, oas, inst} = Workflow.handle_message(wf, inst, msg)","ref":"Toolbox.Workflow.html#module-example"},{"type":"function","title":"Toolbox.Workflow.add_transition/2","doc":"Defines a new transition in a given workflow.\n\nThe transition is defined using a keyword list with the following keys:\n- `from` - source status\n  - required parameter\n- `to` - target status\n  - required parameter\n- `when` - predicate used to select transition which will be executed\n  - optional parameter\n  - possible `when` definitions:\n    - `{Module, function}`, where function accepts transition, instance and message as args,\n      and returns a boolean.\n    - `{:timeout, timeout}`, where timeout is defined in milliseconds\n    - `{:=, [path, to, state, key], value}`\n    - `{:>, [path, to, state, key], value}`\n    - `{:<, [path, to, state, key], value}`\n    - `{:<=, [path, to, state, key], value}`\n    - `{:>=, [path, to, state, key], value}`\n    - `{:contains, [path, to, state, key], value}`\n    - `{:is_in, [path, to, state, key], list_value}`\n    - a list of predicates where each has one of the shapes above\n      - e.g., `[{__MODULE__, :play_msg}, {:=, [:cd, :artist], \"Pink Floyd\"}]`\n      - all predicates in that list must hold for the combined predicate to hold\n- `then` - callback used to update workflow instance *state* during transition execution\n  - optional parameter\n  - possible then definitions:\n    - `{Module, function}`, where function accepts transition, instance and message as args\n    - a list of `{Module, functions}` items\n      - listed callbacks are executed in the given order\n  - calbacks should return `{:ok, wf_instance_state :: map()}` or `{:error, reason}`\n- `side_effects` - callback used to generate output actions during transition execution\n  - optional parameter\n  - possible `side_effects` definitions:\n    - `{Module, function}`, where function accepts transition, instance and message as args\n    - a list of `{Module, functions}` items\n      - listed callbacks are executed in the given order\n  - calbacks should return `{:ok, [Runbox.Scenario.OutputAction.oa_params()]}` or `{:error, reason}`\n- `update_history_entry` - callback used to modify transition execution history entry\n  - optional parameter\n  - possible `update_history_entry` definitions:\n    - `{Module, function}`, where function accepts history entry, transition, instance and message\n      as args\n    - a list of `{Module, functions}` items\n      - listed callbacks are executed in the given order\n  - calbacks should return `{:ok, updated_history_entry :: map()}` or `{:error, reason}`\n- `update_possible_transition` - callback used in `handle_message/3` to modify possible transition\n  - optional parameter\n  - possible then definitions:\n    - `{Module, function}`, where function accepts possible transition, transition, instance\n      and message as args\n    - a list of `{Module, functions}` items\n      - listed callbacks are executed in the given order\n  - calbacks should return `{:ok, updated_possible_transition :: map()}` or `{:error, reason}`","ref":"Toolbox.Workflow.html#add_transition/2"},{"type":"function","title":"Toolbox.Workflow.build/1","doc":"Finishes workflow definition, validates all configured dependencies and workflow structure.","ref":"Toolbox.Workflow.html#build/1"},{"type":"function","title":"Toolbox.Workflow.handle_message/3","doc":"Uses given workflow definition and message to update the status and state of a given instance.\n\nIf no configured workflow transition matches, nothing will happen, i.e., instance status and\nstate will remain the same.\n\nOrder of callback execution:\n1. `when` definitions of transitions in definition order\n2. `then` definitions of matching transition\n3. `update_history_entry` definitions of matching transition\n3. `update_possible_transition` definitions of matching transition\n5. `side_effects` definitions of matching transition","ref":"Toolbox.Workflow.html#handle_message/3"},{"type":"function","title":"Toolbox.Workflow.new/0","doc":"Creates new blank workflow definition.\n\nContinue with calls to `add_transition/2` and `build/1` to define a workflow.","ref":"Toolbox.Workflow.html#new/0"},{"type":"function","title":"Toolbox.Workflow.new_instance/7","doc":"Creates a new instance for a given workflow.\n\n`status` is the initial status of the instance. `state` is its initial state.\n\n`type` and `id` may be arbitrary. If your instance represents an asset or incident,\nyou may use its type and id. These two fields are kept in the `Toolbox.Workflow.Instance`\nstruct and are available in all workflow callbacks, so you can use them e.g. to construct\noutput actions. They don't directly influence the computation, they're just additional\ninformation available to the user.\n\n`msg` is the message that causes this instance to be created. Its timestamp is used as\nthe first timestamp in the instance history.\n\nWhen a new instance is created, it internally performs a transition from the internal\nstatus `:none` to the given `status`. You may pass transition callbacks using `options`\n(in the same way as to `add_transition/2`) and these callbacks will be executed on this\ninitial transition. All the callbacks except for `when` are supported.\n\nThe creation of an instance automatically adds the first entry to the instance history\nand sets the `last_update` field to the timestamp of `msg`.","ref":"Toolbox.Workflow.html#new_instance/7"},{"type":"function","title":"Example - Toolbox.Workflow.new_instance/7","doc":"iex> Workflow.new_instance(wf, \"empty\", \"cd_player\", \"Sony D-50\", %{cd: nil},\n    ...>                       %Message{type: :discover_player, timestamp: 100})\n    {:ok, [],\n      %Toolbox.Workflow.Instance{\n        id: \"Sony D-50\",\n        type: \"cd_player\",\n        state: %{cd: nil},\n        last_update: 100,\n        status: \"empty\",\n        history: [%{\"status\" => \"empty\", \"timestamp\" => 100}],\n        possible_transitions: [%{\"status\" => \"idle\", \"timestamp\" => -1}],\n        next_possible_transition_timestamp: nil,\n        terminated?: false\n    }}","ref":"Toolbox.Workflow.html#new_instance/7-example"},{"type":"type","title":"Toolbox.Workflow.status/0","doc":"","ref":"Toolbox.Workflow.html#t:status/0"},{"type":"type","title":"Toolbox.Workflow.t/0","doc":"","ref":"Toolbox.Workflow.html#t:t/0"},{"type":"type","title":"Toolbox.Workflow.transitions/0","doc":"","ref":"Toolbox.Workflow.html#t:transitions/0"},{"type":"module","title":"Toolbox.Workflow.Instance","doc":"Struct holding the current state of a single workflow instance.\n\nThe struct should be built only via `Toolbox.Workflow.new_instance/7`. An updated instance\nis subsequently returned from every call to `Toolbox.Workflow.handle_message/3`.","ref":"Toolbox.Workflow.Instance.html"},{"type":"type","title":"Toolbox.Workflow.Instance.t/0","doc":"Workflow instance.\n\n* `status` - the current status. Note that the `:none` status may be set only during\n  internal initialization of the instance. After`Toolbox.Workflow.new_instance/7` returns,\n  it will never have this status.\n* `state` - the current context. May be any map. Updated using the `then` callbacks.\n* `type`: an arbitrary “type”, possibly an asset type. Does not influence anything.\n  Set using `Toolbox.Workflow.new_instance/7`, does not change afterwards.\n* `id`: an arbitrary “id”, possibly an asset id. May be used to distinguish instances.\n  Does not influence anything. Set using `Toolbox.Workflow.new_instance/7`, does not\n  change afterwards.\n* `history`: a list of history entries, sorted from oldest to newest. Each entry is by default\n  of the form `%{\"status\" => some_status, \"timestamp\" => some_ts}`. Even the initial status is\n  included.  In case of a timeout transition, the timestamp is the timestamp of the relevant\n  message (i.e., may be later than the actual timeout). The shape of history entries may be\n  modified using the `update_history_entry` callbacks.\n* `last_updated`: the timestamp of the last transition. Including the transition to the\n  initial state. In case of a timeout transition, the timestamp corresponds to the moment\n  when the timeout fires.\n* `possible_transitions`: a list of transitions than can be performed from the current status.\n  Each is by default of the form `%{\"status\" => some_status, \"timestamp\" => some_ts}`.\n  The timestamp is `-1` if there is no timeout bound with the transition, otherwise it is\n  timestamp of the timeout. If a transition uses the builtin `when` conditions on state,\n  they are taken into account and only if they hold, the transition is considered as possible.\n  The shape of the possible transitions may be modified using the `update_possible_transition`\n  callbacks.\n* `next_possible_transition_timestamp`: the earliest timestamp out of the possible transitions\n  that specify a timeout (i.e., their timeout is not `-1`), or `nil` if there is no such\n  transition.\n* `terminated?`: a flag meaning that the instance cannot move to any other status, i.e.,\n  its status is terminal and there are no transition from it.","ref":"Toolbox.Workflow.Instance.html#t:t/0"},{"type":"type","title":"Example - Toolbox.Workflow.Instance.t/0","doc":"This example uses the `CdPlayerWorkflow` from `Toolbox.Workflow` moduledoc.\n\n    %Toolbox.Workflow.Instance{\n      id: \"Sony D-50\",\n      type: \"cd_player\",\n      state: %{cd: %{album: \"Zorya\", artist: \"Floex\"}},\n      last_update: 300,\n      status: \"playing\",\n      history: [\n        %{\"status\" => \"empty\", \"timestamp\" => 100},\n        %{\"status\" => \"idle\", \"timestamp\" => 200},\n        %{\"status\" => \"playing\", \"timestamp\" => 300}\n      ],\n      possible_transitions: [\n        %{\"status\" => \"idle\", \"timestamp\" => -1},\n        %{\"status\" => \"idle\", \"timestamp\" => 4440300},\n        %{\"status\" => \"empty\", \"timestamp\" => -1}\n      ],\n      next_possible_transition_timestamp: 4440300,\n      terminated?: false\n    }","ref":"Toolbox.Workflow.Instance.html#t:t/0-example"},{"type":"module","title":"Toolbox.Workflow.Transition","doc":"Struct defining a single transition, i.e., a change from one status to another.\n\nThe struct should be built only via `Toolbox.Workflow.add_transition/2`. You can then access\nthis struct in all the workflow callbacks such as `when` or `then`.","ref":"Toolbox.Workflow.Transition.html"},{"type":"type","title":"Toolbox.Workflow.Transition.builtin_when_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:builtin_when_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.side_effects_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:side_effects_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.t/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:t/0"},{"type":"type","title":"Toolbox.Workflow.Transition.then_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:then_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.update_history_entry_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:update_history_entry_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.update_possible_transition_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:update_possible_transition_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.when_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:when_fn_def/0"},{"type":"extras","title":"Altworx Toolbox","doc":"# Altworx Toolbox\n\n**Toolbox** is a convenience library designed to streamline the creation of Altworx scenarios.\nIt provides a variety of utilities to simplify tasks like managing incidents, using randomness, producing reports, and more.\nWhile scenarios can benefit from the functionality Toolbox offers, its use is entirely optional and not a requirement for\nrunning scenarios within Altworx.","ref":"readme.html"},{"type":"extras","title":"Installation - Altworx Toolbox","doc":"Add Toolbox as a dependency in your `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:toolbox, \"~> 5.3\", hex: :altworx_toolbox}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Modules overview - Altworx Toolbox","doc":"* `Toolbox.Workflow`: general-purpose state machine abstraction\n* `Toolbox.Incident`: state machine abstraction for managing incidents\n* `Toolbox.Incident.IdGenerator`: helper for generating stable incident IDs\n* `Toolbox.Random`: helpers for using reproducible randomness inside scenarios\n* `Toolbox.Report`: behaviour for sending periodic reports via notifications\n* `Toolbox.Utils.Enum` and `Toolbox.Utils.Map`: utility functions for working with enumerables and maps","ref":"readme.html#modules-overview"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"Master - Changelog","doc":"","ref":"changelog.html#master"},{"type":"extras","title":"Changes - Changelog","doc":"","ref":"changelog.html#changes"},{"type":"extras","title":"5.3.4 - Changelog","doc":"","ref":"changelog.html#5-3-4"},{"type":"extras","title":"Changes - Changelog","doc":"- Bump Runbox to 12.0.0","ref":"changelog.html#changes"},{"type":"extras","title":"5.3.3 - Changelog","doc":"","ref":"changelog.html#5-3-3"},{"type":"extras","title":"Changes - Changelog","doc":"- Bump Runbox to 11.0.0","ref":"changelog.html#changes"},{"type":"extras","title":"5.3.2 - Changelog","doc":"","ref":"changelog.html#5-3-2"},{"type":"extras","title":"Changes - Changelog","doc":"- Improve the documentation\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/27),\n   [#6195](https://altworx.plan.io/issues/6195))\n- Bump runbox version to 10.0.0\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/28),\n   [#5918](https://altworx.plan.io/issues/5918))","ref":"changelog.html#changes"},{"type":"extras","title":"5.3.1 - Changelog","doc":"","ref":"changelog.html#5-3-1"},{"type":"extras","title":"Changes - Changelog","doc":"- Wider support for incident type in incident workflows\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/23),\n   [#6193](https://altworx.plan.io/issues/6193))\n- Improve documentation for `Toolbox.Workflow` and `Toolbox.Incident`\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/23))\n- Bump runbox version to 9.0.0\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/26))","ref":"changelog.html#changes"},{"type":"extras","title":"5.3.0 - Changelog","doc":"","ref":"changelog.html#5-3-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Support additional event parameters in Incident workflow\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/24),\n   [#6592](https://altworx.plan.io/issues/6592))","ref":"changelog.html#changes"},{"type":"extras","title":"5.2.0 - Changelog","doc":"","ref":"changelog.html#5-2-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Add new function to generate incident ID and deprecate the old one.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/21),\n   [#6391](https://altworx.plan.io/issues/6391))","ref":"changelog.html#changes"},{"type":"extras","title":"5.1.1 - Changelog","doc":"","ref":"changelog.html#5-1-1"},{"type":"extras","title":"Changes - Changelog","doc":"- Fix Incident ID Generator on new OTP\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/20))","ref":"changelog.html#changes"},{"type":"extras","title":"5.1.0 - Changelog","doc":"","ref":"changelog.html#5-1-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Depend on Runbox v7 and upgrade Elixir and OTP\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/19),\n   [#6285](https://altworx.plan.io/issues/6285))","ref":"changelog.html#changes"},{"type":"extras","title":"5.0.1 - Changelog","doc":"","ref":"changelog.html#5-0-1"},{"type":"extras","title":"Changes - Changelog","doc":"- Improved documentation of `Toolbox.Incident.add_transition/2`\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/18),\n   [#6117](https://altworx.plan.io/issues/6117))","ref":"changelog.html#changes"},{"type":"extras","title":"Changes - Changelog","doc":"","ref":"changelog.html#changes"},{"type":"extras","title":"5.0.0 - Changelog","doc":"","ref":"changelog.html#5-0-0"},{"type":"extras","title":"Breaking changes - Changelog","doc":"- Expected return value has changed for callback `:side_effects` in\n  `Toolbox.Incident` and `Toolbox.Workflow`. When returning an output action,\n  it should be of type `Runbox.Scenario.OutputAction.oa_params()` (previously\n  `Runbox.Scenario.OutputAction.t()`). See runbox changelog for the version\n  6.0.0.","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Changes - Changelog","doc":"- Bump runbox to 6.0.0. Adapt to new output actions.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/16),\n   [#6177](https://altworx.plan.io/issues/6177))","ref":"changelog.html#changes"},{"type":"extras","title":"4.0.1 - Changelog","doc":"","ref":"changelog.html#4-0-1"},{"type":"extras","title":"Breaking changes - Changelog","doc":"- `Toolbox.Incident.new_instance/7` and `Toolbox.Workflow.new_instance/7` replaced `Toolbox.Incident.new_instance/6` and `Toolbox.Workflow.new_instance/6`. Update scenarios to use the new functions with `type` as an extra argument (which is placed before `id`).\n\n  **Before**:\n  ```elixir\n  Incident.new_instance(\n    incident_def,\n    \"status\",\n    \"/incident/high_temperature/189d359b-686e-36e3-a529-b253706a40df\",\n    %{},\n    msg,\n    severity: 12\n  )\n  ```\n\n  **After**:\n  ```elixir\n  Incident.new_instance(\n    incident_def,\n    \"status\",\n    \"high_temperature\",\n    \"189d359b-686e-36e3-a529-b253706a40df\",\n    %{},\n    msg,\n    severity: 12\n  )\n  ```\n\n  `id` was `/incident/high_temperature/189d359b-686e-36e3-a529-b253706a40df`, is now split to `type` as `high_temperature` and `id` as `189d359b-686e-36e3-a529-b253706a40df`, prefix `/incident/` is omitted because is no longer needed.\n\n- `Toolbox.Incident.new_instance/7`'s parameter `edges` changed to `actors`. In scenarios use `actors` with new format.\n\n  **Before**:\n  ```elixir\n  [%{type: \"actor\", actor_id: \"/assets/security_tools/pam\"}]\n  ```\n  or\n  ```elixir\n  [%{actor_id: \"/assets/security_tools/pam\"}]\n  ```\n\n  **After**:\n  ```elixir\n  [%{type: \"/assets/security_tools\", id: \"pam\"}]\n  ```\n\n  Note the `actor_id` was split into `type` and `id`.\n\n- Interpolation of `{{id}}` in `Toolbox.Incident` attributes has changed.\n  Previously it resolved to the full id, such as\n  `/incident/high_temperature/189d359b-686e-36e3-a529-b253706a40df`. Now it\n  resolves only to the id, such as `189d359b-686e-36e3-a529-b253706a40df`.","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Changes - Changelog","doc":"- Change param edges to actors\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/12)\n   [#6069](https://altworx.plan.io/issues/6069))\n- Split asset ID in Toolbox.Workflow and Toolbox.Incident\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/13)\n   [#6069](https://altworx.plan.io/issues/6067))","ref":"changelog.html#changes"},{"type":"extras","title":"3.1.0 - Changelog","doc":"","ref":"changelog.html#3-1-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Use incident output actions in Incident.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/10)\n   [#5999](https://altworx.plan.io/issues/5999))\n- Default Incident subject and description.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/11)\n   [#6026](https://altworx.plan.io/issues/6026))\n- Fix passing incident history to output action.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/11)\n   [#6026](https://altworx.plan.io/issues/6026))","ref":"changelog.html#changes"},{"type":"extras","title":"3.0.0 - Changelog","doc":"","ref":"changelog.html#3-0-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Move certain modules from Toolbox to Runbox\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/9),\n   [#5848](https://altworx.plan.io/issues/5848))\n\n#### Breaking changes\n- Many modules were moved from Toolbox to Runbox (some kept their name, but the majority\n  of them was also renamed to be under the Runbox \"namespace\"); here are the ones that\n  affect scenario authors or deployment maintainers (for a complete list, see the task):\n  - Modules that will require changes in aliases (just replace `Toolbox` with `Runbox`)\n    - `Toolbox.Notifications.TemplateHelper` → `Runbox.Notifications.TemplateHelper`\n    - `Toolbox.Runtime.RuntimeInstruction` → `Runbox.Runtime.RuntimeInstruction`\n    - `Toolbox.Utils.Path` → `Runbox.Utils.Path`\n    - `Toolbox.Scenario.Config` → `Runbox.Scenario.Config`\n    - `Toolbox.Scenario.Helper` → `Runbox.Scenario.Helper`\n    - `Toolbox.Scenario.Manifest` → `Runbox.Scenario.Manifest`\n  - A module that requires change of `@behaviour` (just replace `Toolbox` with `Runbox`)\n    - `Toolbox.Scenario.Template.StageBased`→ `Runbox.Scenario.Template.StageBased`\n  - Configuration key move\n    - `:scenario_config_dir` config key was moved from `:toolbox` to `:runbox`,\n      you may have to change this in your `config/test.exs`\n  - A module that will require change in pipeline definitions inside `config.ini`\n    - `Toolbox.Normalizer.LoadTopic.Pipeline.parse_msg` → `Normalizer.LoadTopic.Pipeline.parse_msg`","ref":"changelog.html#changes"},{"type":"extras","title":"2.0.0 - Changelog","doc":"","ref":"changelog.html#2-0-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Remove serialization vector from output actions\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/6),\n   [#3788](https://altworx.plan.io/issues/3788))\n\n#### Breaking changes\n- Field `svector` was removed from `Toolbox.Scenario.OutputAction`\n- Function `Toolbox.Scenario.OutputAction.new_create_edge_action/5`\n  was removed, scenarios should be updated to use `Toolbox.Scenario.OutputAction.new_create_edge_action/4`\n- Function `Toolbox.Scenario.OutputAction.new_upsert_edge_action/5`\n  was removed, scenarios should be updated to use `Toolbox.Scenario.OutputAction.new_upsert_edge_action/4`\n- Function `Toolbox.Scenario.OutputAction.new_delete_edge_action/5`\n  was removed, scenarios should be updated to use `Toolbox.Scenario.OutputAction.new_delete_edge_action/4`","ref":"changelog.html#changes"},{"type":"extras","title":"1.1.0 - Changelog","doc":"","ref":"changelog.html#1-1-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Remove redundant API for registering timeouts with both timeout timestamp and timeout message.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/5),\n   [#5827](https://altworx.plan.io/issues/5827))\n\n#### Breaking changes\n- API function to register timeouts `Toolbox.Runtime.RuntimeInstruction.register_timeout/2`\n  was removed, update scenarios to use `Toolbox.Runtime.RuntimeInstruction.register_timeout/1`\n  instead.\n- Field `timeout` was removed from `Toolbox.Runtime.RuntimeInstruction.Timeout`","ref":"changelog.html#changes"},{"type":"extras","title":"1.0.2 - Changelog","doc":"","ref":"changelog.html#1-0-2"},{"type":"extras","title":"Changes - Changelog","doc":"- Remove unused dependency and update others\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/4),\n   [#5661](https://altworx.plan.io/issues/5661))","ref":"changelog.html#changes"},{"type":"extras","title":"1.0.1 - Changelog","doc":"","ref":"changelog.html#1-0-1"},{"type":"extras","title":"Changes - Changelog","doc":"- Explicitly load modules when checking if a function is exported\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/3),\n   [#5784](https://altworx.plan.io/issues/5784))","ref":"changelog.html#changes"},{"type":"extras","title":"1.0.0 - Changelog","doc":"","ref":"changelog.html#1-0-0"},{"type":"extras","title":"Changes - Changelog","doc":"- Add required topics to Scenario Manifest\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/2),\n   [#5725](https://altworx.plan.io/issues/5725))","ref":"changelog.html#changes"},{"type":"extras","title":"0.1.0 - Changelog","doc":"","ref":"changelog.html#0-1-0"},{"type":"extras","title":"Changes - Changelog","doc":"The initial release, moved Toolbox to separate repository.\n([PR](https://gitlab.dev.altworx.com/erlang/product/toolbox/-/merge_requests/1),\n [#5712](https://altworx.plan.io/issues/5712))","ref":"changelog.html#changes"}],"producer":{"name":"ex_doc","version":[48,46,51,52,46,50]},"content_type":"text/markdown"}